#include "combinator.hh"
#include "operator.hh"
#include "timers.hh"
#include "adlib.hh"

Adlib::Adlib()
{
	Timer = new Timers();
	Operators = new Operator[36];
	Combiners = new Combinator[18];
	
	Remap();
}

Adlib::~Adlib()
{
	delete Timer;
	delete[] Operators;
	delete[] Combiners;
}

void Adlib::PairLink(unsigned pairno, unsigned combno, unsigned opno)
{
	if(opl3 && pairs[pairno])
	{
		OPL3_link(Operators[opno],
		          Operators[opno+3],
		          Operators[opno+6],
		          Operators[opno+9],
		          
		          /* FIXME: Check if the order is here correct */		          
		          (Combiners[combno  ].is_am() ? 1 : 0) +
		          (Combiners[combno+3].is_am() ? 2 : 0)
		         );
	}
	else
	{
		OPL2_link(Operators[opno],
		          Operators[opno+3],
		          Combiners[combno].is_am() ? 1 : 0
		         );

		OPL2_link(Operators[opno+6],
		          Operators[opno+9],
		          Combiners[combno+3].is_am() ? 1 : 0
		         );
	}
}

void Adlib::Remap()
{
	PairLink(0, 0, 0);
	PairLink(1, 1, 1);
	PairLink(2, 2, 2);
	PairLink(3, 9,18);
	PairLink(4,10,19);
	PairLink(5,11,20);

	OPL2_link(Operators[12], Operators[15], Combiners[6].is_am() ? 1 : 0);
	OPL2_link(Operators[13], Operators[16], Combiners[7].is_am() ? 1 : 0);
	OPL2_link(Operators[14], Operators[17], Combiners[8].is_am() ? 1 : 0);
	
	OPL2_link(Operators[30], Operators[33], Combiners[15].is_am() ? 1 : 0);
	OPL2_link(Operators[31], Operators[34], Combiners[16].is_am() ? 1 : 0);
	OPL2_link(Operators[32], Operators[35], Combiners[17].is_am() ? 1 : 0);
}

//#include <cstdio>
//using std::printf;

void Adlib::Out(unsigned ind, unsigned char data)
{
	if(ind==0 || ind==2)
		index = data;
	else if(ind == 1 || ind == 3)
	{
		//printf("Adlib: [%03Xh] <- %02Xh\n", index, data);
		
		unsigned char ind8 = index & 0xFF;
	
		if(ind==3)index |= 0x100;
		else index &= 0xFF;
		
		switch(index)
		{
			case 0x002: Timer->WriteTimer1(data); return;
			case 0x003: Timer->WriteTimer2(data); return;
			case 0x004: Timer->WriteReg4(data); return;
			case 0x105:
			    opl3 = data & 1;
			    Remap();
			    return;
			case 0x104:
			    pairs[0] = data&1;
			    pairs[1] = data&2;
			    pairs[2] = data&4;
			    pairs[3] = data&8;
			    pairs[4] = data&16;
			    pairs[5] = data&32;
			    Remap();
			    return;
		}
		
		if(ind8 >= 0x20 && ind8 <= 0x35)
			Operators[PortOffsetOperator(index)].Write20(data);
		else if(ind8 >= 0x40 && ind8 <= 0x55)
			Operators[PortOffsetOperator(index)].Write40(data);
		else if(ind8 >= 0x60 && ind8 <= 0x75)
			Operators[PortOffsetOperator(index)].Write60(data);
		else if(ind8 >= 0x80 && ind8 <= 0x95)
			Operators[PortOffsetOperator(index)].Write80(data);
		else if(ind8 >= 0xA0 && ind8 <= 0xA8)
			Combiners[ind8-0xA0 + 9*(index/256)].WriteA0(data);
		else if(ind8 >= 0xB0 && ind8 <= 0xB8)
			Combiners[ind8-0xA0 + 9*(index/256)].WriteB0(data);
		else if(ind8 >= 0xC0 && ind8 <= 0xC8)
		{
			Combiners[ind8-0xA0 + 9*(index/256)].WriteC0(data);
			Remap();
		}
		else if(ind8 >= 0xE0 && ind8 <= 0xE8)
			Combiners[ind8-0xA0 + 9*(index/256)].WriteE0(data);

		else if(ind8 == 0xBD)
		{
			tremodepth = (data & 0x80) ? dB_4_8 : dB_1_0;
			vibradepth = (data & 0x40) ? cent14 : cent7;
			
			/* UNIMPLEMENTED:
			    5: percussion mode: 
			    0=Hihat,1=Cymbal,2=TomTom,3=Snare,4=Bassdrum
			*/
		}
		else if(ind8 == 0x08)
		{
			/* UNIMPLEMENTED:
			    7 = composite sine wave mode flag
			    6 = keyboard split point flag
			*/
		}
	}
}

unsigned char Adlib::In(unsigned ind)
{
	if(!ind)
	{
		/* One read elapses 0.64 microseconds */
		Run(6.4E-7);
		
		register unsigned char state = Timer->ReadStatus();
		
		//printf("Adlib: State %02Xh\n", state);
		
		return state;
	}
	return 0;
}

void Adlib::Run(double secs)
{
	Timer->Run(secs);
}
